import PropTypes from 'prop-types';
import { createContext, useRef, useCallback, useMemo, useContext, useState, useLayoutEffect, useEffect } from 'react';
import _defineProperty from '@babel/runtime/helpers/defineProperty';
import { jsx } from 'react/jsx-runtime';

let StatusEnum;

(function (StatusEnum) {
  StatusEnum["ERROR"] = "error";
  StatusEnum["IDLE"] = "idle";
  StatusEnum["LOADING"] = "loading";
  StatusEnum["SUCCESS"] = "success";
})(StatusEnum || (StatusEnum = {}));

const KEY_IS_NOT_STRING_ERROR = 'Key should be a string';
const DEFAULT_MAX_CONCURRENT_REQUESTS = 20;

class DataLoader {
  constructor(args) {
    var _this = this;

    _defineProperty(this, "isCalled", false);

    _defineProperty(this, "isCancelled", true);

    _defineProperty(this, "status", StatusEnum.IDLE);

    _defineProperty(this, "observers", []);

    _defineProperty(this, "loadCount", 0);

    _defineProperty(this, "isFirstLoading", true);

    _defineProperty(this, "tryLaunch", async () => {
      if (DataLoader.started < DataLoader.maxConcurrent) {
        DataLoader.started += 1;
        DataLoader.queue[this.key] = this.launch();
      } else {
        DataLoader.queue[this.key] = new Promise(resolve => {
          setTimeout(resolve);
        }).then(this.tryLaunch);
      }

      return DataLoader.queue[this.key];
    });

    _defineProperty(this, "load", async function () {
      let force = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;

      if (force || !_this.isCalled) {
        _this.isCalled = true;

        if (_this.status !== StatusEnum.LOADING) {
          _this.status = StatusEnum.LOADING;

          _this.notifyChanges();
        }

        DataLoader.queue[_this.key] = _this.tryLaunch();
      }

      return DataLoader.queue[_this.key];
    });

    _defineProperty(this, "launch", async () => {
      try {
        this.isCancelled = false;
        this.loadCount += 1;
        const data = await this.method();

        if (!this.isCancelled) {
          DataLoader.cachedData[this.key] = data;
          this.status = StatusEnum.SUCCESS;
          this.data = data;
          this.error = undefined;
        }

        this.isCalled = false;
        this.isFirstLoading = false;
        DataLoader.started -= 1;
        delete DataLoader.queue[this.key];
        this.notifyChanges();
        return data;
      } catch (error) {
        if (!this.isCancelled) {
          this.status = StatusEnum.ERROR;
          this.error = error;
        }

        this.isCalled = false;
        this.isFirstLoading = false;
        DataLoader.started -= 1;
        delete DataLoader.queue[this.key];
        this.notifyChanges();

        if (!this.isCancelled) {
          throw error;
        }

        return undefined;
      }
    });

    this.key = args.key;
    this.method = args.method;

    if (args.enabled) {
      this.status = StatusEnum.LOADING;
    }

    this.data = DataLoader.cachedData[this.key];
    if (args.notifyChanges) this.observers.push(args.notifyChanges);
  }

  notifyChanges() {
    if (this.timeout) {
      clearTimeout(this.timeout);
    }

    this.timeout = setTimeout(() => {
      this.observers.forEach(observer => observer());
    });
  }

  getData() {
    return DataLoader.cachedData[this.key];
  }

  clearData() {
    DataLoader.cachedData[this.key] = undefined;
  }

  cancel() {
    DataLoader.started -= 1;
    delete DataLoader.queue[this.key];
    this.isCancelled = true;
    this.status = StatusEnum.IDLE;
  }

  addObserver(observer) {
    this.observers.push(observer);
  }

  removeObserver(observer) {
    const index = this.observers.indexOf(observer);
    if (index > -1) this.observers.splice(index, 1);
  }

}

_defineProperty(DataLoader, "maxConcurrent", DEFAULT_MAX_CONCURRENT_REQUESTS);

_defineProperty(DataLoader, "started", 0);

_defineProperty(DataLoader, "cachedData", {});

_defineProperty(DataLoader, "queue", {});

const DataLoaderContext = createContext(undefined);

const DataLoaderProvider = _ref => {
  let {
    children,
    cacheKeyPrefix,
    onError,
    maxConcurrentRequests
  } = _ref;
  const requestsRef = useRef({});
  const computeKey = useCallback(key => `${cacheKeyPrefix ? `${cacheKeyPrefix}-` : ''}${key}`, [cacheKeyPrefix]);
  const getRequest = useCallback(key => requestsRef.current[computeKey(key)], [computeKey]);
  const addRequest = useCallback((key, args) => {
    if (DataLoader.maxConcurrent !== maxConcurrentRequests) {
      DataLoader.maxConcurrent = maxConcurrentRequests;
    }

    if (key && typeof key === 'string') {
      const newRequest = new DataLoader({ ...args,
        key: computeKey(key)
      });
      requestsRef.current[newRequest.key] = newRequest;
      return newRequest;
    }

    throw new Error(KEY_IS_NOT_STRING_ERROR);
  }, [computeKey, maxConcurrentRequests]);
  const getOrAddRequest = useCallback((key, args) => {
    const requestFound = getRequest(key);

    if (!requestFound) {
      return addRequest(key, args);
    }

    return requestFound;
  }, [addRequest, getRequest]);
  const clearCachedData = useCallback(key => {
    if (typeof key === 'string') {
      if (requestsRef.current[computeKey(key)]) {
        requestsRef.current[computeKey(key)].clearData();
      }
    } else throw new Error(KEY_IS_NOT_STRING_ERROR);
  }, [computeKey]);
  const clearAllCachedData = useCallback(() => {
    Object.values(requestsRef.current).forEach(request => {
      request.clearData();
    });
  }, []);
  const reload = useCallback(async key => {
    if (key && typeof key === 'string') {
      var _getRequest;

      await ((_getRequest = getRequest(key)) === null || _getRequest === void 0 ? void 0 : _getRequest.load(true));
    } else throw new Error(KEY_IS_NOT_STRING_ERROR);
  }, [getRequest]);
  const reloadAll = useCallback(async () => {
    await Promise.all(Object.values(requestsRef.current).map(request => request.load(true)));
  }, []);
  const getCachedData = useCallback(key => {
    if (key) {
      var _getRequest2;

      return (_getRequest2 = getRequest(key)) === null || _getRequest2 === void 0 ? void 0 : _getRequest2.getData();
    }

    return Object.values(requestsRef.current).reduce((acc, request) => ({ ...acc,
      [request.key]: request.getData()
    }), {});
  }, [getRequest]);
  const getReloads = useCallback(key => {
    if (key) {
      return getRequest(key) ? () => getRequest(key).load(true) : undefined;
    }

    return Object.entries(requestsRef.current).reduce((acc, _ref2) => {
      let [requestKey, {
        load
      }] = _ref2;
      return { ...acc,
        [requestKey]: () => load(true)
      };
    }, {});
  }, [getRequest]);
  const value = useMemo(() => ({
    addRequest,
    cacheKeyPrefix,
    clearAllCachedData,
    clearCachedData,
    getCachedData,
    getOrAddRequest,
    getReloads,
    getRequest,
    onError,
    reload,
    reloadAll
  }), [addRequest, cacheKeyPrefix, clearAllCachedData, clearCachedData, getCachedData, getOrAddRequest, getRequest, getReloads, onError, reload, reloadAll]);
  return jsx(DataLoaderContext.Provider, {
    value: value,
    children: children
  });
};

DataLoaderProvider.propTypes = {
  cacheKeyPrefix: PropTypes.string,
  children: PropTypes.node.isRequired,
  maxConcurrentRequests: PropTypes.number,
  onError: PropTypes.func
};
DataLoaderProvider.defaultProps = {
  cacheKeyPrefix: undefined,
  maxConcurrentRequests: DEFAULT_MAX_CONCURRENT_REQUESTS,
  onError: undefined
};
const useDataLoaderContext = () => useContext(DataLoaderContext);

function useDataLoader(fetchKey, method) {
  let {
    enabled = true,
    onError,
    onSuccess,
    keepPreviousData = false,
    needPolling = true,
    pollingInterval,
    initialData
  } = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  const {
    getOrAddRequest,
    onError: onGlobalError
  } = useDataLoaderContext();
  const methodRef = useRef(method);
  const onSuccessRef = useRef(onSuccess);
  const onErrorRef = useRef(onError ?? onGlobalError);
  const isMountedRef = useRef(false);
  const [, setCounter] = useState(0);
  const forceRerender = useCallback(() => {
    setCounter(current => current + 1);
  }, []);
  useLayoutEffect(() => {
    isMountedRef.current = true;
    return () => {
      isMountedRef.current = false;
    };
  });
  const request = getOrAddRequest(fetchKey, {
    enabled,
    method: methodRef.current
  });
  useEffect(() => {
    request.addObserver(forceRerender);
    return () => {
      request.removeObserver(forceRerender);
    };
  }, [request, forceRerender]);
  const previousDataRef = useRef(request.data);
  const isLoading = request.status === StatusEnum.LOADING;
  const isSuccess = request.status === StatusEnum.SUCCESS;
  const isError = request.status === StatusEnum.ERROR;
  const isIdle = request.status === StatusEnum.IDLE && !enabled;
  const isPolling = !!(pollingInterval && (typeof needPolling === 'function' && (request.isFirstLoading || needPolling(request.data)) || typeof needPolling !== 'function' && needPolling));
  const reload = useCallback(() => request.load(true).then(onSuccessRef.current).catch(onErrorRef.current), [request]);
  useEffect(() => {
    request.method = method;
  }, [method, request]);
  useEffect(() => {
    onSuccessRef.current = onSuccess;
  }, [onSuccess]);
  useEffect(() => {
    onErrorRef.current = onError ?? onGlobalError;
  }, [onError, onGlobalError]);
  useEffect(() => {
    if (keepPreviousData) {
      previousDataRef.current = request.data;
    }
  }, [request.data, keepPreviousData]);
  useEffect(() => {
    if (enabled && !request.isCalled) {
      request.load().then(onSuccessRef.current).catch(onErrorRef.current);
    }
  }, [enabled, request, keepPreviousData]);
  useEffect(() => {
    let timeout;

    if (isMountedRef.current && request && pollingInterval && needPolling && (request.status === StatusEnum.SUCCESS || request.status === StatusEnum.ERROR)) {
      if (typeof needPolling === 'function' && needPolling(request.data) || typeof needPolling !== 'function' && needPolling) {
        timeout = setTimeout(() => {
          request.load(true).then(onSuccessRef.current).catch(onErrorRef.current);
        }, pollingInterval);
      }
    }

    return () => {
      if (timeout && !isMountedRef.current) clearTimeout(timeout);
    };
  }, [pollingInterval, needPolling, request, request.status, request.data]);
  return {
    data: !request.isFirstLoading ? request.data : initialData,
    error: request.error,
    isError,
    isIdle,
    isLoading,
    isPolling,
    isSuccess,
    previousData: previousDataRef.current,
    reload
  };
}

const usePaginatedDataLoader = function (baseFetchKey, method) {
  let {
    enabled = true,
    initialData,
    keepPreviousData = true,
    onError,
    onSuccess,
    pollingInterval,
    maxDataLifetime,
    needPolling,
    initialPage,
    perPage = 1
  } = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  if (typeof baseFetchKey !== 'string') {
    throw new Error(KEY_IS_NOT_STRING_ERROR);
  }

  const [data, setData] = useState({});
  const [page, setPage] = useState(initialPage ?? 1);
  const pageMethod = useCallback(() => method({
    page,
    perPage
  }), [method, page, perPage]);
  const {
    data: pageData,
    isError,
    isIdle,
    isLoading,
    isPolling,
    isSuccess,
    reload,
    error
  } = useDataLoader(`${baseFetchKey}-page-${page}`, pageMethod, {
    enabled,
    initialData,
    keepPreviousData,
    maxDataLifetime,
    needPolling,
    onError,
    onSuccess,
    pollingInterval
  });
  const goToNextPage = useCallback(() => {
    setPage(current => current + 1);
  }, []);
  const goToPreviousPage = useCallback(() => {
    setPage(current => current > 1 ? current - 1 : 1);
  }, []);
  const goToPage = useCallback(newPage => {
    setPage(newPage > 1 ? newPage : 1);
  }, []);
  useEffect(() => {
    setData(current => {
      if (pageData !== current[page]) {
        return { ...current,
          [page]: pageData
        };
      }

      return current;
    });
  }, [pageData, page]);
  useEffect(() => {
    setPage(1);
    setData({});
  }, [baseFetchKey]);
  return {
    data,
    error,
    goToNextPage,
    goToPage,
    goToPreviousPage,
    isError,
    isIdle,
    isLoading,
    isPolling,
    isSuccess,
    page,
    pageData,
    reload
  };
};

export { DataLoaderProvider, useDataLoader, useDataLoaderContext, usePaginatedDataLoader };
