import { ReactNode, ReactElement } from 'react';
import PropTypes from 'prop-types';

declare enum StatusEnum {
    ERROR = "error",
    IDLE = "idle",
    LOADING = "loading",
    SUCCESS = "success"
}

declare class PromiseType<T = unknown> extends Promise<T> {
    cancel?: () => void;
}
declare type OnErrorFn = ((err: Error) => void | Promise<void>) | undefined;
declare type OnSuccessFn<T = unknown> = ((result: T) => void | Promise<void>) | undefined;
declare type NeedPollingType<T = unknown> = boolean | ((data?: T) => boolean);
/**
 * @typedef {Object} UseDataLoaderConfig
 * @property {Function} [onSuccess] callback when a request success
 * @property {Function} [onError] callback when a error is occured, this will override the onError specified on the Provider if any
 * @property {*} [initialData] initial data if no one is present in the cache before the request
 * @property {number} [pollingInterval] relaunch the request after the last success
 * @property {boolean} [enabled=true] launch request automatically (default true)
 * @property {boolean} [keepPreviousData=true] do we need to keep the previous data after reload (default true)
 * @property {NeedPollingType} [needPolling=true] When pollingInterval is set you can set a set a custom callback to know if polling is enabled
 */
interface UseDataLoaderConfig<T = unknown> {
    enabled?: boolean;
    initialData?: T;
    keepPreviousData?: boolean;
    onError?: OnErrorFn;
    onSuccess?: OnSuccessFn;
    pollingInterval?: number;
    /**
     * Max time before data from previous success is considered as outdated (in millisecond)
     */
    maxDataLifetime?: number;
    needPolling?: NeedPollingType<T>;
}
/**
 * @typedef {Object} UseDataLoaderResult
 * @property {boolean} isIdle true if the hook in initial state
 * @property {boolean} isLoading true if the request is launched
 * @property {boolean} isSuccess true if the request success
 * @property {boolean} isError true if the request throw an error
 * @property {boolean} isPolling true if the request if enabled is true, pollingInterval is defined and the status is isLoading or isSuccess
 * @property {*} previousData if keepPreviousData is true it return the last data fetched
 * @property {*} data initialData if no data is fetched or not present in the cache otherwise return the data fetched
 * @property {string} error the error occured during the request
 * @property {Function} reload reload the data
 */
interface UseDataLoaderResult<T = unknown, ErrorType = Error> {
    data?: T;
    error?: ErrorType;
    isError: boolean;
    isIdle: boolean;
    isLoading: boolean;
    isPolling: boolean;
    isSuccess: boolean;
    previousData?: T;
    reload: () => Promise<void>;
}
/**
 * Params send to the method
 */
declare type UsePaginatedDataLoaderMethodParams = {
    page: number;
    perPage: number;
};
declare type UsePaginatedDataLoaderConfig<T = unknown> = UseDataLoaderConfig<T> & {
    initialPage?: number;
    perPage?: number;
};
declare type UsePaginatedDataLoaderResult<T = unknown> = {
    pageData?: T;
    data?: Record<number, T | undefined>;
    error?: Error;
    isError: boolean;
    isIdle: boolean;
    isLoading: boolean;
    isPolling: boolean;
    isSuccess: boolean;
    reload: () => Promise<void>;
    goToPage: (page: number) => void;
    goToNextPage: () => void;
    goToPreviousPage: () => void;
    page: number;
};

declare type DataLoaderConstructorArgs<T = unknown> = {
    key: string;
    method: () => PromiseType<T>;
    enabled?: boolean;
    notifyChanges?: () => void;
};
declare class DataLoader<ResultType = unknown, ErrorType = unknown> {
    static maxConcurrent: number;
    static started: number;
    static cachedData: Record<string, unknown | undefined>;
    static queue: Record<string, PromiseType>;
    key: string;
    method: () => PromiseType<ResultType>;
    isCalled: boolean;
    isCancelled: boolean;
    status: StatusEnum;
    error?: ErrorType;
    data?: ResultType;
    observers: Array<() => void>;
    timeout?: number;
    loadCount: number;
    isFirstLoading: boolean;
    constructor(args: DataLoaderConstructorArgs<ResultType>);
    notifyChanges(): void;
    getData(): ResultType | undefined;
    private tryLaunch;
    load: (force?: boolean) => Promise<ResultType>;
    launch: () => Promise<ResultType | undefined>;
    clearData(): void;
    cancel(): void;
    addObserver(observer: () => void): void;
    removeObserver(observer: () => void): void;
}

declare type CachedData = Record<string, unknown>;
declare type Reloads = Record<string, () => Promise<void | unknown>>;
declare type UseDataLoaderInitializerArgs<T = unknown> = {
    method: () => PromiseType<T>;
    /**
     * Max time before data from previous success is considered as outdated (in millisecond)
     */
    maxDataLifetime?: number;
    enabled?: boolean;
};
declare type GetCachedDataFn = {
    (): CachedData;
    (key?: string): unknown | undefined;
};
declare type GetReloadsFn = {
    (): Reloads;
    (key?: string): (() => Promise<void | unknown>) | undefined;
};
interface IDataLoaderContext {
    addRequest: (key: string, args: UseDataLoaderInitializerArgs) => DataLoader;
    getOrAddRequest: <T>(key: string, args: UseDataLoaderInitializerArgs<T>) => DataLoader<T>;
    cacheKeyPrefix?: string;
    onError?: (error: Error) => void | Promise<void>;
    clearAllCachedData: () => void;
    clearCachedData: (key: string) => void;
    getCachedData: GetCachedDataFn;
    getReloads: GetReloadsFn;
    getRequest: (key: string) => DataLoader;
    reload: (key?: string) => Promise<void>;
    reloadAll: () => Promise<void>;
}
declare const DataLoaderProvider: {
    ({ children, cacheKeyPrefix, onError, maxConcurrentRequests, }: {
        children: ReactNode;
        cacheKeyPrefix: string;
        onError: OnErrorFn;
        maxConcurrentRequests?: number | undefined;
    }): ReactElement;
    propTypes: {
        cacheKeyPrefix: PropTypes.Requireable<string>;
        children: PropTypes.Validator<string | number | boolean | PropTypes.ReactElementLike | PropTypes.ReactNodeArray>;
        maxConcurrentRequests: PropTypes.Requireable<number>;
        onError: PropTypes.Requireable<(...args: any[]) => any>;
    };
    defaultProps: {
        cacheKeyPrefix: undefined;
        maxConcurrentRequests: number;
        onError: undefined;
    };
};
declare const useDataLoaderContext: () => IDataLoaderContext;

declare function useDataLoader<ResultType, ErrorType = Error>(fetchKey: string, method: () => PromiseType<ResultType>, { enabled, onError, onSuccess, keepPreviousData, needPolling, pollingInterval, initialData, }?: UseDataLoaderConfig<ResultType>): UseDataLoaderResult<ResultType, ErrorType>;

/**
 * @param {string} baseFetchKey base key used to cache data. Hook append -page-X to that key for each page you load
 * @param {() => PromiseType} method a method that return a promise
 * @param {useDataLoaderConfig} config hook configuration
 * @returns {useDataLoaderResult} hook result containing data, request state, and method to reload the data
 */
declare const usePaginatedDataLoader: <T>(baseFetchKey: string, method: (params: UsePaginatedDataLoaderMethodParams) => PromiseType<T>, { enabled, initialData, keepPreviousData, onError, onSuccess, pollingInterval, maxDataLifetime, needPolling, initialPage, perPage, }?: UsePaginatedDataLoaderConfig<T>) => UsePaginatedDataLoaderResult<T>;

export { DataLoaderProvider, useDataLoader, useDataLoaderContext, usePaginatedDataLoader };
